/* ====================================================================
 * Copyright (c) 1995-1998 The Apache Group.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer. 
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. All advertising materials mentioning features or use of this
 *    software must display the following acknowledgment:
 *    "This product includes software developed by the Apache Group
 *    for use in the Apache HTTP server project (http://www.apache.org/)."
 *
 * 4. The names "Apache Server" and "Apache Group" must not be used to
 *    endorse or promote products derived from this software without
 *    prior written permission. For written permission, please contact
 *    apache@apache.org.
 *
 * 5. Products derived from this software may not be called "Apache"
 *    nor may "Apache" appear in their names without prior written
 *    permission of the Apache Group.
 *
 * 6. Redistributions of any form whatsoever must retain the following
 *    acknowledgment:
 *    "This product includes software developed by the Apache Group
 *    for use in the Apache HTTP server project (http://www.apache.org/)."
 *
 * THIS SOFTWARE IS PROVIDED BY THE APACHE GROUP ``AS IS'' AND ANY
 * EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE APACHE GROUP OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 * ====================================================================
 *
 * This software consists of voluntary contributions made by many
 * individuals on behalf of the Apache Group and was originally based
 * on public domain software written at the National Center for
 * Supercomputing Applications, University of Illinois, Urbana-Champaign.
 * For more information on the Apache Group and the Apache HTTP server
 * project, please see <http://www.apache.org/>.
 *
 */

/*
 * http_script: keeps all script-related ramblings together.
 * 
 * Compliant to CGI/1.1 spec
 * 
 * Adapted by rst from original NCSA code by Rob McCool
 *
 * Apache adds some new env vars; REDIRECT_URL and REDIRECT_QUERY_STRING for
 * custom error responses, and DOCUMENT_ROOT because we found it useful.
 * It also adds SERVER_ADMIN - useful for scripts to know who to mail when 
 * they fail.
 */

#include "httpd.h"
#include "http_config.h"
#include "http_request.h"
#include "http_core.h"
#include "http_protocol.h"
#include "http_main.h"
#include "http_log.h"
#include "util_script.h"
#include "http_conf_globals.h"

module MODULE_VAR_EXPORT cgi_module;

/* XXX: need to re-implement ScriptLog -- should use layered i/o */

/* KLUDGE --- for back-combatibility, we don't have to check ExecCGI
 * in ScriptAliased directories, which means we need to know if this
 * request came through ScriptAlias or not... so the Alias module
 * leaves a note for us.
 */

static int is_scriptaliased(request_rec *r)
{
    const char *t = ap_table_get(r->notes, "alias-forced-type");
    return t && (!strcasecmp(t, "cgi-script"));
}

/* Configuration stuff */

#define DEFAULT_LOGBYTES 10385760
#define DEFAULT_BUFBYTES 1024

typedef struct {
    char *logname;
    long logbytes;
    int bufbytes;
} cgi_server_conf;

static void *create_cgi_config(pool *p, server_rec *s)
{
    cgi_server_conf *c =
    (cgi_server_conf *) ap_pcalloc(p, sizeof(cgi_server_conf));

    c->logname = NULL;
    c->logbytes = DEFAULT_LOGBYTES;
    c->bufbytes = DEFAULT_BUFBYTES;

    return c;
}

static void *merge_cgi_config(pool *p, void *basev, void *overridesv)
{
    cgi_server_conf *base = (cgi_server_conf *) basev, *overrides = (cgi_server_conf *) overridesv;

    return overrides->logname ? overrides : base;
}

static const char *set_scriptlog(cmd_parms *cmd, void *dummy, char *arg)
{
    server_rec *s = cmd->server;
    cgi_server_conf *conf =
    (cgi_server_conf *) ap_get_module_config(s->module_config, &cgi_module);

    conf->logname = arg;
    return NULL;
}

static const char *set_scriptlog_length(cmd_parms *cmd, void *dummy, char *arg)
{
    server_rec *s = cmd->server;
    cgi_server_conf *conf =
    (cgi_server_conf *) ap_get_module_config(s->module_config, &cgi_module);

    conf->logbytes = atol(arg);
    return NULL;
}

static const char *set_scriptlog_buffer(cmd_parms *cmd, void *dummy, char *arg)
{
    server_rec *s = cmd->server;
    cgi_server_conf *conf =
    (cgi_server_conf *) ap_get_module_config(s->module_config, &cgi_module);

    conf->bufbytes = atoi(arg);
    return NULL;
}

static const command_rec cgi_cmds[] =
{
    {"ScriptLog", set_scriptlog, NULL, RSRC_CONF, TAKE1,
     "the name of a log for script debugging info"},
    {"ScriptLogLength", set_scriptlog_length, NULL, RSRC_CONF, TAKE1,
     "the maximum length (in bytes) of the script debug log"},
    {"ScriptLogBuffer", set_scriptlog_buffer, NULL, RSRC_CONF, TAKE1,
     "the maximum size (in bytes) to record of a POST request"},
    {NULL}
};

static int log_scripterror(request_rec *r, int ret,
			   int show_errno, char *error)
{
    BUFF *f;
    PRFileInfo finfo;
    char timebuf[MAX_STRING_LEN];
    cgi_server_conf *conf;
    
    ap_log_error(APLOG_MARK, show_errno|APLOG_ERR, r->server, 
		"%s: %s", error, r->filename);

    conf = ap_get_module_config(r->server->module_config, &cgi_module);
    if (!conf->logname ||
	((PR_GetFileInfo(ap_server_root_relative(r->pool, conf->logname), &finfo) == PR_SUCCESS)
	 &&   (finfo.size > conf->logbytes)) ||
         ((f = ap_bopenf(r->pool, ap_server_root_relative(r->pool, conf->logname),
		      PR_WRONLY|PR_APPEND|PR_CREATE_FILE, 0666)) == NULL)) {
	return ret;
    }

    /* "%% [Wed Jun 19 10:53:21 1996] GET /cgi-bin/printenv HTTP/1.0" */
    ap_bprintf(f, "%%%% [%s] %s %s%s%s %s\n", ap_get_time(timebuf), r->method, r->uri,
	    r->args ? "?" : "", r->args ? r->args : "", r->protocol);
    /* "%% 500 /usr/local/apache/cgi-bin */
    ap_bprintf(f, "%%%% %d %s\n", ret, r->filename);

    ap_bprintf(f, "%%error\n%s\n", error);

    ap_bclose(f);
    return ret;
}

#if 0
static int log_script(request_rec *r, cgi_server_conf * conf, int ret,
		  char *dbuf, const char *sbuf, BUFF *script_in, BUFF *script_err)
{
    array_header *hdrs_arr = table_elts(r->headers_in);
    table_entry *hdrs = (table_entry *) hdrs_arr->elts;
    char argsbuffer[HUGE_STRING_LEN];
    BUFF *f;
    int i;
    PRFileInfo finfo;

    if (!conf->logname ||
	((PR_GetFileInfo(ap_server_root_relative(r->pool, conf->logname), &finfo) == PR_SUCCESS)
	 &&   (finfo.size > conf->logbytes)) ||
         ((f = ap_bopenf(r->pool, ap_server_root_relative(r->pool, conf->logname),
		      PR_WRONLY|PR_APPEND|PR_CREATE_FILE, 0666)) == NULL)) {
	/* Soak up script output */
	while (ap_bgets(argsbuffer, sizeof(argsbuffer), script_in) > 0)
	    continue;
	while (ap_bgets(argsbuffer, sizeof(argsbuffer), script_err) > 0)
	    continue;
	return ret;
    }

    /* "%% [Wed Jun 19 10:53:21 1996] GET /cgi-bin/printenv HTTP/1.0" */
    ap_bprintf(f, "%%%% [%s] %s %s%s%s %s\n", ap_get_time(argsbuffer), r->method, r->uri,
	    r->args ? "?" : "", r->args ? r->args : "", r->protocol);
    /* "%% 500 /usr/local/apache/cgi-bin" */
    ap_bprintf(f, "%%%% %d %s\n", ret, r->filename);

    ap_bputs("%request\n", f);
    for (i = 0; i < hdrs_arr->nelts; ++i) {
	if (!hdrs[i].key)
	    continue;
	ap_bprintf(f, "%s: %s\n", hdrs[i].key, hdrs[i].val);
    }
    if ((r->method_number == M_POST || r->method_number == M_PUT)
	&& *dbuf) {
	ap_bprintf(f, "\n%s\n", dbuf);
    }

    ap_bputs("%response\n", f);
    hdrs_arr = table_elts(r->err_headers_out);
    hdrs = (table_entry *) hdrs_arr->elts;

    for (i = 0; i < hdrs_arr->nelts; ++i) {
	if (!hdrs[i].key)
	    continue;
	ap_bprintf(f, "%s: %s\n", hdrs[i].key, hdrs[i].val);
    }

    if (sbuf && *sbuf)
	ap_bprintf(f, "%s\n", sbuf);

    if (ap_bgets(argsbuffer, sizeof(argsbuffer), script_in) > 0) {
	ap_bputs("%stdout\n", f);
	ap_bputs(argsbuffer, f);
	while (ap_bgets(argsbuffer, sizeof(argsbuffer), script_in) > 0)
	    ap_bputs(argsbuffer, f);
	ap_bputs("\n", f);
    }

    if (ap_bgets(argsbuffer, sizeof(argsbuffer), script_err) > 0) {
	ap_bputs("%stderr\n", f);
	ap_bputs(argsbuffer, f);
	while (ap_bgets(argsbuffer, sizeof(argsbuffer), script_err) > 0)
	    ap_bputs(argsbuffer, f);
	ap_bputs("\n", f);
    }

    ap_bclose(script_in);
    ap_bclose(script_err);

    ap_bclose(f);
    return ret;
}
#endif

/****************************************************************
 *
 * Actual CGI handling...
 */

static int cgi_handler(request_rec *r)
{
    int retval, nph;
    char *argv0;
    BUFF *script_out, *script_in;
    char argsbuffer[HUGE_STRING_LEN];
    int is_included = !strcmp(r->protocol, "INCLUDED");
    PRFileDesc *child_in, *parent_out;
    PRFileDesc *child_out, *parent_in;
    PRProcessAttr *attr;
    PRProcess *pid;
    char **env;

    if (r->method_number == M_OPTIONS) {
	/* 99 out of 100 CGI scripts, this is all they support */
	r->allowed |= (1 << M_GET);
	r->allowed |= (1 << M_POST);
	return DECLINED;
    }

    if ((argv0 = strrchr(r->filename, '/')) != NULL)
	argv0++;
    else
	argv0 = r->filename;

    nph = !(strncmp(argv0, "nph-", 4));

    if (!(ap_allow_options(r) & OPT_EXECCGI) && !is_scriptaliased(r))
	return log_scripterror(r, FORBIDDEN, APLOG_NOERRNO,
			       "Options ExecCGI is off in this directory");
    if (nph && is_included)
	return log_scripterror(r, FORBIDDEN, APLOG_NOERRNO,
			       "attempt to include NPH CGI script");

#if defined(__EMX__) || defined(WIN32)
    /* Allow for cgi files without the .EXE extension on them under OS/2 */
    if (!r->file_exists) {
	PRFileInfo finfo;

	r->filename = ap_pstrcat(r->pool, r->filename, ".EXE", NULL);

	if (PR_GetFileInfo(r->filename, &finfo) != 0
	    || finfo.type != PR_FILE_FILE) {
	    return log_scripterror(r, NOT_FOUND, 0,
				   "script not found or unable to stat");
	}
    }
#else
    if (!r->file_exists)
	return log_scripterror(r, NOT_FOUND, APLOG_NOERRNO,
			       "script not found or unable to stat");
#endif
    if (r->finfo.type == PR_FILE_DIRECTORY)
	return log_scripterror(r, FORBIDDEN, APLOG_NOERRNO,
			       "attempt to invoke directory as script");

    if ((retval = ap_setup_client_block(r, REQUEST_CHUNKED_ERROR)))
	return retval;

    ap_add_common_vars(r);
    ap_add_cgi_vars(r);
    env = ap_create_environment(r->pool, r->subprocess_env);

#ifdef CHARSET_EBCDIC
    /* XXX:@@@ Is the generated/included output ALWAYS in text/ebcdic format? */
    /* Or must we check the Content-Type first? */
    ap_bsetflag(r->connection->client, B_EBCDIC2ASCII, 1);
#endif /*CHARSET_EBCDIC*/

    if (PR_CreatePipe(&child_in, &parent_out) != PR_SUCCESS) {
	ap_log_error(APLOG_MARK, APLOG_ERR, r->server,
	    "unable to create pipe for child process: %s", r->filename);
	goto setup_bail_0;
    }

    if (PR_CreatePipe(&parent_in, &child_out) != PR_SUCCESS) {
	ap_log_error(APLOG_MARK, APLOG_ERR, r->server,
	    "unable to create pipe for child process: %s", r->filename);
	goto setup_bail_1;
    }

    attr = PR_NewProcessAttr();
    PR_SetStdioRedirect(attr, PR_StandardInput, child_in);
    PR_SetStdioRedirect(attr, PR_StandardOutput, child_out);
    PR_SetStdioRedirect(attr, PR_StandardError, r->server->error_log);
    /* XXX: this is where we would tell NSPR to do a chdir() in the child */

    pid = ap_spawn_cgi(r, argv0, env, 0, attr);
    if (pid == NULL) {
	ap_log_error(APLOG_MARK, APLOG_ERR, r->server,
	    "unable to create child process: %s", r->filename);
	PR_DestroyProcessAttr(attr);
	PR_Close(child_out);
	PR_Close(parent_in);
setup_bail_1:
	PR_Close(child_in);
	PR_Close(parent_out);
setup_bail_0:
	return HTTP_INTERNAL_SERVER_ERROR;
    }

    PR_Close(child_in);
    PR_Close(child_out);
    PR_DestroyProcessAttr(attr);

    /* We detach it, and don't even bother noting it in any pool...
     * because we're going to close stdin and stdout and the script
     * will get EOF when it tries to read or write.  A broken script
     * could go forever.  But that's a broken script.  We can deal with
     * those later.
     */
    if (PR_DetachProcess(pid) != PR_SUCCESS) {
	ap_log_error(APLOG_MARK, APLOG_INFO, r->server,
	    "(ignored) error detaching process: %s", r->filename);
    }

    /* XXX: this is where we'd set up timeouts for reads/writes to the CGI,
     * but there's no way to do this in NSPR currently.
     */
    script_in = ap_bcreate(r->pool, B_RD);
    ap_bpushfd(script_in, parent_in);
    script_out = ap_bcreate(r->pool, B_WR);
    ap_bpushfd(script_out, parent_out);

    /* Transfer any put/post args, CERN style... */
    if (ap_should_client_block(r)) {
	int len_read;

	while ((len_read =
		ap_get_client_block(r, argsbuffer, sizeof(argsbuffer))) > 0) {
	    if (ap_bwrite(script_out, argsbuffer, len_read) < len_read) {
		/* silly script stopped reading, soak up remaining message */
		while (ap_get_client_block(r, argsbuffer, sizeof(argsbuffer)) > 0) {
		    /* dump it */
		}
		break;
	    }
	}

	ap_bflush(script_out);
    }

    ap_bclose(script_out);

    /* Handle script return... */
    if (!nph) {
	const char *location;
	int ret;

	if ((ret = ap_scan_script_header_err(r, script_in))) {
	    return ret;
	}

#ifdef CHARSET_EBCDIC
        /* Now check the Content-Type to decide if conversion is needed */
        ap_checkconv(r);
#endif /*CHARSET_EBCDIC*/

	location = ap_table_get(r->headers_out, "Location");
	if (location && location[0] == '/' && r->status == 200) {
	    /* Soak up all the script output */
	    while (ap_bgets(argsbuffer, sizeof(argsbuffer), script_in) > 0) {
		continue;
	    }

	    /* This redirect needs to be a GET no matter what the original
	     * method was.
	     */
	    r->method = ap_pstrdup(r->pool, "GET");
	    r->method_number = M_GET;

	    /* We already read the message body (if any), so don't allow
	     * the redirected request to think it has one.  We can ignore 
	     * Transfer-Encoding, since we used REQUEST_CHUNKED_ERROR.
	     */
	    ap_table_unset(r->headers_in, "Content-Length");

	    ap_internal_redirect_handler(location, r);
	    return OK;
	}
	else if (location && r->status == 200) {
	    /* XX Note that if a script wants to produce its own Redirect
	     * body, it now has to explicitly *say* "Status: 302"
	     */
	    return REDIRECT;
	}

	ap_send_http_header(r);
	if (!r->header_only) {
	    ap_send_fb(script_in, r);
	}
    }
    else {
	ap_send_fb(script_in, r);
    }
    ap_bclose(script_in);

    return OK;			/* NOT r->status, even if it has changed. */
}

static const handler_rec cgi_handlers[] =
{
    {CGI_MAGIC_TYPE, cgi_handler},
    {"cgi-script", cgi_handler},
    {NULL}
};

module MODULE_VAR_EXPORT cgi_module =
{
    STANDARD_MODULE_STUFF,
    NULL,			/* initializer */
    NULL,			/* dir config creater */
    NULL,			/* dir merger --- default is to override */
    create_cgi_config,		/* server config */
    merge_cgi_config,		/* merge server config */
    cgi_cmds,			/* command table */
    cgi_handlers,		/* handlers */
    NULL,			/* filename translation */
    NULL,			/* check_user_id */
    NULL,			/* check auth */
    NULL,			/* check access */
    NULL,			/* type_checker */
    NULL,			/* fixups */
    NULL,			/* logger */
    NULL,			/* header parser */
    NULL,			/* child_init */
    NULL,			/* child_exit */
    NULL			/* post read-request */
};
